@@ -1,36 +1,97 @@
-from __future__ import annotations
+# Copyright (c) QuantCo and pydiverse contributors 2025-2025
+# SPDX-License-Identifier: BSD-3-Clause
 
+import dataclasses
 import functools
 import pickle
 import time
+import traceback
 import uuid
 from collections import defaultdict
+from concurrent.futures import Future, ThreadPoolExecutor
 from contextlib import contextmanager
 from contextvars import ContextVar
 from enum import Enum
-from threading import Lock
+from threading import Lock, RLock
 from typing import TYPE_CHECKING, Any
 
+import attrs
 import msgpack
 import structlog
 
+from pydiverse.common.util import Disposable
+from pydiverse.pipedag._typing import T
 from pydiverse.pipedag.context.context import (
+    BaseAttrsContext,
     BaseContext,
     ConfigContext,
     StageLockContext,
 )
 from pydiverse.pipedag.errors import LockError, RemoteProcessError, StageError
-from pydiverse.pipedag.util import Disposable
 from pydiverse.pipedag.util.ipc import IPCServer
 
 if TYPE_CHECKING:
-    from pydiverse.pipedag import Flow, Stage
-    from pydiverse.pipedag._typing import T
     from pydiverse.pipedag.backend import BaseLockManager, LockState
-    from pydiverse.pipedag.core import Task
+    from pydiverse.pipedag.context.trace_hook import TraceHook
+    from pydiverse.pipedag.core import Flow, Stage, Subflow, Task
     from pydiverse.pipedag.materialize import Blob, Table
 
 
+# States
+
+
+class StageState(Enum):
+    UNINITIALIZED = 0
+    INITIALIZING = 1
+    READY = 2
+    COMMITTING = 3
+    COMMITTED = 4
+
+    FAILED = 127
+
+
+class FinalTaskState(Enum):
+    UNKNOWN = 0
+
+    COMPLETED = 1
+    CACHE_VALID = 2
+    FAILED = 10
+
+    def is_successful(self) -> bool:
+        return self in (FinalTaskState.COMPLETED, FinalTaskState.CACHE_VALID)
+
+    def is_failure(self) -> bool:
+        return self in (FinalTaskState.FAILED,)
+
+
+class MemoState(Enum):
+    NONE = 0
+    WAITING = 1
+
+    FAILED = 127
+
+
+# Deferred Table Store operations
+
+
+@dataclasses.dataclass
+class DeferredTableStoreOp:
+    class Condition(Enum):
+        ON_STAGE_CHANGED = 0
+        """Some table produced in the stage is determined to be cache invalid"""
+
+        ON_STAGE_ABORT = 1
+        """The stage has finished running and all tables are still cache valid"""
+
+        ON_STAGE_COMMIT = 2
+        """The stage has finished running and some tables weren't cache valid"""
+
+    fn_name: str
+    condition: Condition
+    args: tuple | list = ()
+    kwargs: dict = dataclasses.field(default_factory=dict)
+
+
 def synchronized(lock_attr: str):
     def decorator(func: T) -> T:
         @functools.wraps(func)
@@ -62,7 +123,7 @@ class RunContextServer(IPCServer):
     and deserialization in case of multi-node execution of tasks in pipedag graph.
     """
 
-    def __init__(self, flow: Flow):
+    def __init__(self, subflow: "Subflow", trace_hook: "TraceHook"):
         config_ctx = ConfigContext.get()
         interface = config_ctx.network_interface
 
@@ -72,7 +133,10 @@ class RunContextServer(IPCServer):
             msg_ext_hook=_msg_ext_hook,
         )
 
-        self.flow = flow
+        self.flow = subflow.flow
+        self.subflow = subflow
+        self.trace_hook = trace_hook
+        self.config_ctx = config_ctx
         self.run_id = uuid.uuid4().hex[:20]
 
         self.stages = list(self.flow.stages.values())
@@ -84,41 +148,61 @@ class RunContextServer(IPCServer):
 
         # STATE
         self.ref_count = [0] * num_stages
+        self.task_state = [FinalTaskState.UNKNOWN] * len(self.tasks)
         self.stage_state = [StageState.UNINITIALIZED] * num_stages
 
         self.table_names = [set() for _ in range(num_stages)]
         self.blob_names = [set() for _ in range(num_stages)]
 
+        self.tables_in_local_table_cache = [set() for _ in range(num_stages)]
+
         self.task_memo: defaultdict[Any, Any] = defaultdict(lambda: MemoState.NONE)
 
+        # DEFERRED TABLE STORE OPERATIONS
+        max_workers = 5
+        if hasattr(config_ctx.store.table_store, "max_concurrent_copy_operations"):
+            max_workers = config_ctx.store.table_store.max_concurrent_copy_operations
+        self.deferred_thread_pool = ThreadPoolExecutor(max_workers=max_workers)
+        self.deferred_ts_ops: dict[int, list[DeferredTableStoreOp]] = {}
+        self.deferred_ts_ops_futures: dict[int, list[Future]] = {}
+        self.changed_stages: set[int] = set()
+
         # STATE LOCKS
         self.ref_count_lock = Lock()
         self.stage_state_lock = Lock()
         self.names_lock = Lock()
         self.task_memo_lock = Lock()
+        self.deferred_ops_lock = RLock()
+        self.local_table_cache_lock = Lock()
 
         # LOCKING
-        config_ctx = ConfigContext.get()
-        self.lock_manager = config_ctx.create_lock_manager()
-        self.lock_handler = StageLockStateHandler(self.lock_manager)
-
         try:
             # If we are inside a StageLockContext, then we shouldn't release any
             # stage locks, instead they get released when StageLockContext.__exit__
             # gets called.
             stage_lock_context = StageLockContext.get()
-            stage_lock_context.lock_state_handlers.append(self.lock_handler)
+            if stage_lock_context.lock_state_handlers:
+                # reuse existing lock manager and handler
+                self.lock_handler = stage_lock_context.lock_state_handlers[0]
+                self.lock_manager = self.lock_handler.lock_manager
+            else:
+                self.lock_manager = config_ctx.create_lock_manager()
+                self.lock_handler = StageLockStateHandler(self.lock_manager)
+                stage_lock_context.lock_state_handlers.append(self.lock_handler)
             self.keep_stages_locked = True
         except LookupError:
             self.keep_stages_locked = False
+            self.lock_manager = config_ctx.create_lock_manager()
+            self.lock_handler = StageLockStateHandler(self.lock_manager)
+        self.trace_hook.run_init_context_server(self)
 
     def __enter__(self):
         super().__enter__()
+        from pydiverse.pipedag.backend.lock import LockState
 
         # INITIALIZE EVERYTHING
         with self.lock_manager("_pipedag_setup_"):
-            config_ctx = ConfigContext.get()
-            config_ctx.store.table_store.setup()
+            self.config_ctx.store.table_store.setup()
 
             # Acquire a lock on all stages
             # We must lock all stages from the start to prevent two flows from
@@ -126,7 +210,10 @@ class RunContextServer(IPCServer):
             # only when it's needed (may result in deadlocks), the lock should
             # get acquired in the `PipeDAGStore.init_stage` function instead.
             for stage in self.stages:
-                self.lock_manager.acquire(stage)
+                # only lock stage if it is not already locked within current
+                # StageLockContext
+                if self.lock_manager.get_lock_state(stage.name) != LockState.LOCKED:
+                    self.lock_manager.acquire(stage.name)
 
         # INITIALIZE REFERENCE COUNTERS
         for stage in self.stages:
@@ -136,9 +223,12 @@ class RunContextServer(IPCServer):
 
         self.__context_proxy = RunContext(self)
         self.__context_proxy.__enter__()
+        self.deferred_thread_pool.__enter__()
         return self.__context_proxy
 
     def __exit__(self, exc_type, exc_val, exc_tb):
+        self.deferred_thread_pool.shutdown(wait=True, cancel_futures=True)
+        self.deferred_thread_pool.__exit__(exc_type, exc_val, exc_tb)
         self.__context_proxy.__exit__(exc_type, exc_val, exc_tb)
 
         if not self.keep_stages_locked:
@@ -146,17 +236,31 @@ class RunContextServer(IPCServer):
 
         super().__exit__(exc_type, exc_val, exc_tb)
 
-    def handle_request(self, request):
+    def handle_request(self, request: dict[str, Any]):
         try:
             op_name, args = request
+            assert isinstance(op_name, str)
             op = getattr(self, op_name)
-            if not callable(op):
-                raise TypeError(f"OP '{op_name}' is not callable")
+            if not callable(op) or op_name[0] == "_":
+                raise TypeError(f"OP '{op_name}' is not allowed on RunContextServer")
 
             result = op(*args)
             return [None, result]
         except Exception as e:
-            pickled_exception = pickle.dumps(e)
+            exception_tb = traceback.format_exc()
+            try:
+                # Add more context to exception
+                exception_with_traceback = Exception(f"{type(e).__name__}: {e}\n{exception_tb}")
+                exception_with_traceback.__cause__ = e
+                pickled_exception = pickle.dumps(exception_with_traceback)
+            except Exception as e2:
+                self.logger.error(
+                    "failed pickling exception",
+                    traceback=exception_tb,
+                    pickle_exception=str(e2),
+                )
+                pickled_exception = pickle.dumps(RuntimeError("failed pickling exception\n" + exception_tb))
+
             return [pickled_exception, None]
 
     # STAGE: Reference Counting
@@ -186,8 +290,18 @@ class RunContextServer(IPCServer):
             StageState.INITIALIZING,
             StageState.READY,
         )
+        self.trace_hook.stage_post_init(stage_id, success)
 
     def enter_commit_stage(self, stage_id: int):
+        self.trace_hook.stage_pre_commit(stage_id)
+
+        self._await_deferred_ts_ops(stage_id)
+        if self.has_stage_changed(stage_id):
+            self._trigger_deferred_ts_ops(stage_id, DeferredTableStoreOp.Condition.ON_STAGE_COMMIT)
+        else:
+            self._trigger_deferred_ts_ops(stage_id, DeferredTableStoreOp.Condition.ON_STAGE_ABORT)
+        self._await_deferred_ts_ops(stage_id)
+
         return self._enter_stage_state_transition(
             stage_id,
             StageState.READY,
@@ -202,6 +316,7 @@ class RunContextServer(IPCServer):
             StageState.COMMITTING,
             StageState.COMMITTED,
         )
+        self.trace_hook.stage_post_commit(stage_id, success)
 
     def _enter_stage_state_transition(
         self,
@@ -220,9 +335,11 @@ class RunContextServer(IPCServer):
                 self.stage_state[stage_id] = transition
                 return True
 
+        self.logger.debug("Waiting for stage transition...")
         # Wait until transition is over
         while self.stage_state[stage_id] == transition:
             time.sleep(0.01)
+        self.logger.debug("Waiting completed")
 
         # Check if transition was successful
         with self.stage_state_lock:
@@ -231,10 +348,7 @@ class RunContextServer(IPCServer):
             if state == to:
                 return False
             if state == StageState.FAILED:
-                raise StageError(
-                    f"Stage '{self.stages[stage_id].name}' is in unexpected state"
-                    f" {state}"
-                )
+                raise StageError(f"Stage '{self.stages[stage_id].name}' is in unexpected state {state}")
 
     def _exit_stage_state_transition(
         self,
@@ -258,28 +372,83 @@ class RunContextServer(IPCServer):
 
     def acquire_stage_lock(self, stage_id: int):
         stage = self.stages[stage_id]
-        self.lock_manager.acquire(stage)
+        self.lock_manager.acquire(stage.name)
 
     def release_stage_lock(self, stage_id: int):
         if not self.keep_stages_locked:
             stage = self.stages[stage_id]
-            self.lock_manager.release(stage)
+            self.lock_manager.release(stage.name)
 
     def validate_stage_lock(self, stage_id: int):
         stage = self.stages[stage_id]
         self.lock_handler.validate_stage_lock(stage)
 
+    # DEFERRED TABLE STORE OPERATIONS
+    # Allows deferring operations that should be run on the table store object
+    # until certain events occur. This is required to implement deferred copying
+    # of tables from one schema to another.
+
+    @synchronized("deferred_ops_lock")
+    def defer_table_store_op(self, stage_id: int, op: DeferredTableStoreOp):
+        if stage_id not in self.deferred_ts_ops:
+            self.deferred_ts_ops[stage_id] = []
+            self.deferred_ts_ops_futures[stage_id] = []
+        self.deferred_ts_ops[stage_id].append(op)
+
+        if self.has_stage_changed(stage_id):
+            self._trigger_deferred_ts_ops(stage_id, DeferredTableStoreOp.Condition.ON_STAGE_CHANGED)
+
+    @synchronized("deferred_ops_lock")
+    def has_stage_changed(self, stage_id: int) -> bool:
+        return stage_id in self.changed_stages
+
+    @synchronized("deferred_ops_lock")
+    def set_stage_has_changed(self, stage_id: int):
+        self.changed_stages.add(stage_id)
+        self._trigger_deferred_ts_ops(stage_id, DeferredTableStoreOp.Condition.ON_STAGE_CHANGED)
+
+    @synchronized("deferred_ops_lock")
+    def _trigger_deferred_ts_ops(self, stage_id: int, condition: DeferredTableStoreOp.Condition):
+        # Partition deferred ops into those that should and shouldn't get executed
+        deferred_ts_ops = self.deferred_ts_ops.get(stage_id, [])
+
+        ops_to_execute = [op for op in deferred_ts_ops if op.condition == condition]
+        remaining_ops = [op for op in deferred_ts_ops if op.condition != condition]
+
+        if len(ops_to_execute) == 0:
+            return
+
+        # Trigger ops and store futures
+        store = self.config_ctx.store
+        for op in ops_to_execute:
+            fn = getattr(store.table_store, op.fn_name)
+            assert callable(fn)
+
+            run_context = self.__context_proxy
+            future = self.deferred_thread_pool.submit(_call_with_args, fn, run_context, op.args, op.kwargs)
+            self.deferred_ts_ops_futures[stage_id].append(future)
+
+        # Remove ops that just have been executed
+        self.deferred_ts_ops[stage_id] = remaining_ops
+
+    @synchronized("deferred_ops_lock")
+    def _await_deferred_ts_ops(self, stage_id: int):
+        if stage_id not in self.deferred_ts_ops_futures:
+            return
+
+        futures = self.deferred_ts_ops_futures[stage_id]
+        for future in futures:
+            future.result()
+
     # TASK
 
     def did_finish_task(self, task_id: int, final_state_value: int):
-        # TODO: Do something with the final state.
-        #       For example: The object returned by flow.run could have a list
-        #       of tasks and their final states.
-        _final_state = FinalTaskState(final_state_value)
+        final_state = FinalTaskState(final_state_value)
         task = self.tasks[task_id]
 
-        stages_to_release = []
+        self.task_state[task_id] = final_state
 
+        stages_to_release = []
         with self.ref_count_lock:
             for stage in task.upstream_stages:
                 self.ref_count[stage.id] -= 1
@@ -296,7 +465,10 @@ class RunContextServer(IPCServer):
 
         if not self.keep_stages_locked:
             for stage in stages_to_release:
-                self.lock_manager.release(stage)
+                self.lock_manager.release(stage.name)
+
+    def get_task_states(self) -> list[int]:
+        return [state.value for state in self.task_state]
 
     def enter_task_memo(self, task_id: int, cache_key: str) -> tuple[bool, Any]:
         task = self.tasks[task_id]
@@ -310,10 +482,7 @@ class RunContextServer(IPCServer):
 
         if memo is MemoState.WAITING:
             self.logger.info(
-                (
-                    "Task is currently being run with the same inputs."
-                    " Waiting for the other task to finish..."
-                ),
+                ("Task is currently being run with the same inputs. Waiting for the other task to finish..."),
                 task=task,
             )
         while memo is MemoState.WAITING:
@@ -389,6 +558,25 @@ class RunContextServer(IPCServer):
         for blob in blobs:
             self.blob_names[stage_id].remove(blob)
 
+    # LOCAL TABLE CACHE
+
+    @synchronized("local_table_cache_lock")
+    def should_store_table_in_cache(self, stage_id: int, table_name: str) -> bool:
+        if table_name in self.tables_in_local_table_cache[stage_id]:
+            # Some other task already stored / is attempting to store this table
+            # In case we are running tasks in parallel and `store_input` is set
+            # to true, it can happen that multiple tasks are reading and writing
+            # the same table at the same time, which may lead to incorrect behaviour.
+            return False
+
+        self.tables_in_local_table_cache[stage_id].add(table_name)
+        return True
+
+
+def _call_with_args(fn, run_context, args, kwargs):
+    with run_context:
+        fn(*args, **kwargs)
+
 
 class RunContext(BaseContext):
     _context_var = ContextVar("run_context_proxy")
@@ -397,9 +585,8 @@ class RunContext(BaseContext):
         self.client = server.get_client()
         self.flow = server.flow
         self.run_id = server.run_id
-
-        # Configuration Options
-        self.ignore_fresh_input = False
+        self.trace_hook = server.trace_hook
+        self.trace_hook.run_init_context(self)
 
     def _request(self, op: str, *args):
         error, result = self.client.request([op, args])
@@ -410,17 +597,17 @@ class RunContext(BaseContext):
 
     # STAGE: Reference Counting
 
-    def get_stage_ref_count(self, stage: Stage) -> int:
+    def get_stage_ref_count(self, stage: "Stage") -> int:
         return self._request("get_stage_ref_count", stage.id)
 
     # STAGE: State
 
-    def get_stage_state(self, stage: Stage) -> StageState:
+    def get_stage_state(self, stage: "Stage") -> StageState:
         value = self._request("get_stage_state", stage.id)
         return StageState(value)
 
     @contextmanager
-    def init_stage(self, stage: Stage):
+    def init_stage(self, stage: "Stage"):
         can_continue = self._request("enter_init_stage", stage.id)
         if not can_continue:
             yield False
@@ -436,7 +623,7 @@ class RunContext(BaseContext):
             self._request("exit_init_stage", stage.id, True)
 
     @contextmanager
-    def commit_stage(self, stage: Stage):
+    def commit_stage(self, stage: "Stage"):
         can_continue = self._request("enter_commit_stage", stage.id)
         if not can_continue:
             yield False
@@ -453,22 +640,40 @@ class RunContext(BaseContext):
 
     # STAGE: Lock
 
-    def acquire_stage_lock(self, stage: Stage):
+    def acquire_stage_lock(self, stage: "Stage"):
         self._request("acquire_stage_lock", stage.id)
 
-    def release_stage_lock(self, stage: Stage):
+    def release_stage_lock(self, stage: "Stage"):
         self._request("release_stage_lock", stage.id)
 
-    def validate_stage_lock(self, stage: Stage):
+    def validate_stage_lock(self, stage: "Stage"):
         self._request("validate_stage_lock", stage.id)
 
+    # DEFERRED TABLE STORE OPERATIONS
+
+    def defer_table_store_op(self, stage: "Stage", op: DeferredTableStoreOp):
+        """Defer running a table store operation until a specific stage event"""
+        return self._request("defer_table_store_op", stage.id, op)
+
+    def has_stage_changed(self, stage: "Stage") -> bool:
+        """Check if a stage has changed and still is 100% cache valid"""
+        return self._request("has_stage_changed", stage.id)
+
+    def set_stage_has_changed(self, stage: "Stage"):
+        """Inform the run context that a stage isn't 100% cache valid anymore"""
+        return self._request("set_stage_has_changed", stage.id)
+
     # TASK
 
-    def did_finish_task(self, task: Task, final_state: FinalTaskState):
+    def did_finish_task(self, task: "Task", final_state: FinalTaskState):
         self._request("did_finish_task", task.id, final_state.value)
 
+    def get_task_states(self) -> dict["Task", FinalTaskState]:
+        states = [FinalTaskState(value) for value in self._request("get_task_states")]
+        return {task: states[task.id] for task in self.flow.tasks}
+
     @contextmanager
-    def task_memo(self, task: Task, cache_key: str):
+    def task_memo(self, task: "Task", cache_key: str):
         success, memo = self._request("enter_task_memo", task.id, cache_key)
 
         try:
@@ -479,18 +684,19 @@ class RunContext(BaseContext):
         else:
             self._request("exit_task_memo", task.id, cache_key, True)
 
-    def store_task_memo(self, task: Task, cache_key: str, result: Any):
+    def store_task_memo(self, task: "Task", cache_key: str, result: Any):
         self._request("store_task_memo", task.id, cache_key, result)
 
     # TABLE / BLOB: Names
 
     def add_names(
-        self, stage: Stage, tables: list[Table], blobs: list[Blob]
+        self, stage: "Stage", tables: list["Table"], blobs: list["Blob"]
     ) -> tuple[bool, tuple[str, ...], tuple[str, ...]]:
         response = self._request(
             "add_names",
             stage.id,
-            [t.name for t in tables],
+            # ExternalTableReferences don't create a table in schema of stage
+            [t.name for t in tables if t.external_schema is None],
             [b.name for b in blobs],
         )
 
@@ -499,16 +705,23 @@ class RunContext(BaseContext):
         else:
             return False, response["table_duplicates"], response["blob_duplicates"]
 
-    def remove_names(self, stage: Stage, tables: list[Table], blobs: list[Blob]):
+    def remove_names(self, stage: "Stage", tables: list["Table"], blobs: list["Blob"]):
         self._request(
             "remove_names",
             stage.id,
-            [t.name for t in tables],
+            # ExternalTableReferences don't create a table in schema of stage
+            [t.name for t in tables if t.external_schema is None],
             [b.name for b in blobs],
         )
 
+    # LOCAL TABLE CACHE
 
-class DematerializeRunContext(BaseContext):
+    def should_store_table_in_cache(self, table: "Table"):
+        return self._request("should_store_table_in_cache", table.stage.id, table.name)
+
+
+@attrs.frozen
+class DematerializeRunContext(BaseAttrsContext):
     """Dummy RunContext that returns `COMMITTED` as the stage state for all stages
 
     To dematerialize an object we must know it the associates state has been
@@ -520,12 +733,19 @@ class DematerializeRunContext(BaseContext):
 
     _context_var = RunContext._context_var
 
-    def get_stage_state(self, stage: Stage) -> StageState:
+    flow: "Flow"
+
+    def get_stage_state(self, stage: "Stage") -> StageState:
+        _ = stage
         return StageState.COMMITTED
 
-    def validate_stage_lock(self, stage: Stage):
+    def validate_stage_lock(self, stage: "Stage"):
         pass
 
+    def should_store_table_in_cache(self, table: "Table"):
+        _ = table
+        return False
+
 
 # Stage Locking
 
@@ -536,8 +756,8 @@ class StageLockStateHandler(Disposable):
     or externally UNLOCKED).
     """
 
-    def __init__(self, lock_manager: BaseLockManager):
-        self.logger = structlog.get_logger(cls=type(self).__name__)
+    def __init__(self, lock_manager: "BaseLockManager"):
+        self.logger = structlog.get_logger(logger_name=type(self).__name__)
         self.lock_manager = lock_manager
         self.lock_manager.add_lock_state_listener(self._lock_state_listener)
 
@@ -546,11 +766,9 @@ class StageLockStateHandler(Disposable):
         self.lock_manager.dispose()
         super().dispose()
 
-    def _lock_state_listener(
-        self, stage: Stage, old_state: LockState, new_state: LockState
-    ):
+    def _lock_state_listener(self, stage: "Stage", old_state: "LockState", new_state: "LockState"):
         """Internal listener that gets notified when the state of a lock changes"""
-        from pydiverse.pipedag.backend import LockState
+        from pydiverse.pipedag.backend.lock import LockState
         from pydiverse.pipedag.core import Stage
 
         if not isinstance(stage, Stage):
@@ -558,14 +776,9 @@ class StageLockStateHandler(Disposable):
 
         # Logging
         if new_state == LockState.UNCERTAIN:
-            self.logger.warning(
-                f"Lock for stage '{stage.name}' transitioned to UNCERTAIN state."
-            )
+            self.logger.warning(f"Lock for stage '{stage.name}' transitioned to UNCERTAIN state.")
         if old_state == LockState.UNCERTAIN and new_state == LockState.LOCKED:
-            self.logger.info(
-                f"Lock for stage '{stage.name}' is still LOCKED (after being"
-                " UNCERTAIN)."
-            )
+            self.logger.info(f"Lock for stage '{stage.name}' is still LOCKED (after being UNCERTAIN).")
         if old_state == LockState.UNCERTAIN and new_state == LockState.INVALID:
             self.logger.error(f"Lock for stage '{stage.name}' has become INVALID.")
 
@@ -574,7 +787,7 @@ class StageLockStateHandler(Disposable):
 
         did_log = False
         while True:
-            state = self.lock_manager.get_lock_state(stage)
+            state = self.lock_manager.get_lock_state(stage.name)
 
             if state == LockState.LOCKED:
                 return
@@ -584,10 +797,7 @@ class StageLockStateHandler(Disposable):
                 raise LockError(f"Lock for stage '{stage.name}' is invalid.")
             elif state == LockState.UNCERTAIN:
                 if not did_log:
-                    self.logger.info(
-                        f"Waiting for stage '{stage.name}' lock state to"
-                        " become known again..."
-                    )
+                    self.logger.info(f"Waiting for stage '{stage.name}' lock state to become known again...")
                     did_log = True
 
                 time.sleep(0.01)
@@ -596,40 +806,18 @@ class StageLockStateHandler(Disposable):
                 raise ValueError(f"Invalid state '{state}'.")
 
 
-# States
-
-
-class StageState(Enum):
-    UNINITIALIZED = 0
-    INITIALIZING = 1
-    READY = 2
-    COMMITTING = 3
-    COMMITTED = 4
-
-    FAILED = 127
-
-
-class FinalTaskState(Enum):
-    UNKNOWN = 0
-
-    COMPLETED = 1
-    FAILED = 2
-    SKIPPED = 3
-
-
-class MemoState(Enum):
-    NONE = 0
-    WAITING = 1
-
-    FAILED = 127
-
-
 # msgpack hooks
 # Only used to transmit Table and Blob type metadata
 
 
 def _msg_default(obj):
-    return msgpack.ExtType(0, pickle.dumps(obj))
+    try:
+        ret = pickle.dumps(obj)
+    except Exception as e:
+        logger = structlog.get_logger(logger_name="RunContext msg_default")
+        logger.exception("failed to pickle object", object=obj)
+        raise e
+    return msgpack.ExtType(0, ret)
 
 
 def _msg_ext_hook(code, data):
